import unittest
from epidemik import EpiModel

class EpiModelTestCase(unittest.TestCase):
	def setUp(self):
		self.SIR = EpiModel()
		self.beta = 0.3
		self.mu = 0.1
		self.SIR.add_interaction('S', 'I', 'I', self.beta)
		self.SIR.add_spontaneous('I', 'R', self.mu)

	def test_R0(self):
		self.assertEqual(self.SIR.R0(), 3.0, 'incorrect R0')

	def test_edges(self):
		edges = list(self.SIR.transitions.edges(data=True))
		self.assertEqual(len(edges), 2)

		for edge in edges:
			data = edge[2]

			if "agent" in data:
				self.assertEqual(data['agent'], 'I')
				self.assertEqual(data['rate'], self.beta)
				self.assertEqual(edge[0], 'S')
				self.assertEqual(edge[1], 'I')
			else:
				self.assertEqual(data['rate'], self.mu)
				self.assertEqual(edge[0], 'I')
				self.assertEqual(edge[1], 'R')



### −∗− mode : python ; −∗−
# @file MetaEpiModel.py
# @author Bruno Goncalves
######################################################

import networkx as nx
import numpy as np
from numpy import linalg
from numpy import random
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from matplotlib.collections import PatchCollection

from .EpiModel import *

from tqdm import tqdm
tqdm.pandas()

class MetaEpiModel:
    """Simple Epidemic Model Implementation
    
        Provides a way to implement and numerically integrate 
    """
    def __init__(self, travel_graph, populations, population='Population'):
        """
        Initialize the EpiModel object
        
        Parameters:
        - compartments_: list of strings, optional
            List of compartment names
        
        Returns:
        None
        """
        self.travel_graph = travel_graph
        self.populations = populations
        self.population = population

        models = {}

        self.transitions = None
        self.prototype = None
        self.seasonality = None

        for state in travel_graph.index:
            models[state] = EpiModel() 
            if self.transitions is None:
                self.transitions = models[state].transitions
                self.prototype = models[state] 

        self.models = models

    def __repr__(self):
        """
        Return a string representation of the EpiModel object
        
        Returns:
        string
            String representation of the EpiModel object
        """

        model_text = self.models[self.travel_graph.index[0]]

        text = "Metapopulation model with %u populations\n\nThe disease is defined by an %s" % (self.travel_graph.shape[0], model_text)
        return text 

    def add_interaction(self, source, target, agent, rate):  
        """
        Add an interaction between two compartments_
        
        Parameters:
        - source: string
            Name of the source compartment
        - target: string
            Name of the target compartment
        - agent: string
            Name of the agent
        - rate: float
            Rate of the interaction
        
        Returns:
        None
        """    
        for state in self.models:  
            self.models[state].add_interaction(source, target, agent, rate)       
        
    def add_spontaneous(self, source, target, rate):
        """
        Add a spontaneous transition between two compartments_
        
        Parameters:
        - source: string
            Name of the source compartment
        - target: string
            Name of the target compartment
        - rate: float
            Rate of the transition
        
        Returns:
        None
        """
        for state in self.models:  
            self.models[state].add_spontaneous(source, target, rate)      

    def add_vaccination(self, source, target, rate, start):
        """
        Add a vaccination transition between two compartments_
        
        Parameters:
        - source: string
            Name of the source compartment
        - target: string
            Name of the target compartment
        - rate: float
            Rate of the vaccination
        - start: int
            Start time of the vaccination
        
        Returns:
        None
        """
        for state in self.models:  
            self.models[state].add_vaccination(source, target, rate, start)
    
    def R0(self):
        key = list(self.models.keys())[0]
        return self.models[key].R0()

    def get_state(self, state):
        """
        Return a reference to a state EpiModel object

        Parameters:
        - state: string
            Name of the state to return
        """

        return self.models[state]

    def _initialize_populations(self, susceptible, population=None):
        columns = list(self.transitions.nodes())
        self.compartments_ = pd.DataFrame(np.zeros((self.travel_graph.shape[0], len(columns)), dtype='int'), columns=columns)
        self.compartments_.index = self.populations.index

        if population is None:
            population = self.population

        for state in self.compartments_.index:
            self.compartments_.loc[state, susceptible]  = self.populations.loc[state, population]

    def _run_travel(self, compartments_, travel):
        def travel_step(x, populations):
            n = populations.loc[x.name]
            p = travel.loc[x.name].values.tolist()
            output = np.random.multinomial(n, p)

            return pd.Series(output, index=travel.columns)
        
        
        new_compartments = compartments_.copy()
        
        # Travel occurs independently for each compartment
        # since we don't allow in-flight transitions
        for comp in compartments_.columns:
            new_compartments[comp] = travel.apply(travel_step, populations=compartments_[comp]).sum(axis=1)
            
        return new_compartments
    
    def _run_spread(self):
        for state in self.compartments_.index:
            pop = self.compartments_.loc[state].to_dict()
            self.models[state].single_step(**pop)
            self.compartments_.loc[state] = self.models[state].values_.iloc[[-1]].values[0]

    def simulate(self, timestamp, t_min=1, seasonality=None, seed_state=None, susceptible='S', **kwargs):
        if seed_state is None:
            raise NotInitialized("You have to specify the seed_state")

        self._initialize_populations(susceptible)

        pos = {comp: i for i, comp in enumerate(self.models[seed_state].transitions.nodes())}

        for comp in kwargs:
            if comp != susceptible and comp in pos:
                self.compartments_.loc[seed_state, comp] += kwargs[comp]
                self.compartments_.loc[seed_state, susceptible] -= kwargs[comp]

        for t in tqdm(range(t_min, timestamp+1), total=timestamp):
            self._run_spread()
            self.compartments_ = self._run_travel(self.compartments_, self.travel_graph)
    
    def integrate(self, **kwargs):
        raise NotImplementedError("MetaEpiModel doesn't support direct integration of the ODE")

    def draw_model(self):
        return self.models.iloc[0].draw_model()

    def plot(self, title=None, normed=True, layout=None, **kwargs):
        if layout is None:
            n_pop = self.travel_graph.shape[0]
            N = int(np.round(np.sqrt(n_pop), 0)+1)

            coords = [[i%N, i//N] for i in range(n_pop)]
            coords = pd.DataFrame(coords, columns=['x', 'y'])
            coords.index = self.travel_graph.index
        else:
            coords = layout

        fig, ax = plt.subplots(1, figsize=(16, 22))
        ax.set_aspect(1.)
        ax.invert_yaxis()

        patches = []
        color_list = []
        colors = plt.rcParams['axes.prop_cycle'].by_key()['color']
        x = np.linspace(0., 0.75, self.models['NY'].values_.shape[0])



        for state in self.models:
            daily = self.models[state]['I'].values
            
            timeline = (daily/daily.max())
            color = colors[3]
            
            peak = self.models[state]['I'][self.models[state]['I']>0].index.min()
            
            if peak <= 10:
                color = colors[0]
            elif peak <= 20:
                color = colors[1]
            elif peak <= 30:
                color = colors[2]
            elif peak <= 40:
                color = colors[3]
            elif peak > 40:
                color = colors[4]
                
            fancybox = mpatches.FancyBboxPatch([coords.x[state]-0.5, coords.y[state]-0.5], 0.8, 0.8,
                                                boxstyle=mpatches.BoxStyle("Round", pad=0.06))
            patches.append(fancybox)
            color_list.append(color)
            
            ax.text(y=coords.y[state]-0.4, x=coords.x[state]-0.4, 
                    s=state, horizontalalignment='center', verticalalignment='center', fontsize=15)
            ax.plot(coords.x[state]+x-0.5, coords.y[state]-timeline/1.4+0.25, lw=1, color=colors[1])
            
            ax.vlines(x=coords.x[state]+x[peak]-0.5, 
                    ymin=coords.y[state]-timeline.min()/1.4+0.25, 
                    ymax=coords.y[state]-timeline.max()/1.4+0.25, lw=1,
                    color='darkgray')

        collection = PatchCollection(patches, facecolors=color_list, alpha=0.3)
        ax.add_collection(collection)

        x_min, x_max = ax.get_xlim()
        y_min, y_max = ax.get_ylim()

        x_delta = (x_max-x_min+1)/7

        ax.text(x=1, y=y_min, s='[0-10]',  color=colors[0], fontsize=15, ha='center')
        ax.text(x=2, y=y_min, s='(10-20]', color=colors[1], fontsize=15, ha='center')
        ax.text(x=3, y=y_min, s='(20-30]', color=colors[2], fontsize=15, ha='center')
        ax.text(x=4, y=y_min, s='(30-40]', color=colors[3], fontsize=15, ha='center')
        ax.text(x=5, y=y_min, s='40+',     color=colors[4], fontsize=15, ha='center')


        ax.axis('off')
        fig.patch.set_facecolor('#FFFFFF')
        fig.tight_layout()

    def plot_peaks(self):
        peaks = None

        for state in self.models.values():
            if peaks is None:
                peaks = state.values_[['I']].copy()
            else:
                peaks = pd.concat([peaks, state.values_[['I']]], axis=1)

        peaks = peaks.T
        peaks.index = self.travel_graph.index
        peaks = peaks.apply(lambda x:x/x.max(), axis=1)

        n, m = peaks.shape
        ratio = int(np.round(m/n, 0)+1)
        fig, ax = plt.subplots(1, figsize=(15, 10*ratio))
        import matplotlib as mpl

        norm = mpl.colors.Normalize(vmin=0,vmax=1)
        sm = plt.cm.ScalarMappable(cmap=plt.cm.rainbow, norm=norm)
        sm.set_array([])
        cbar = plt.colorbar(sm, ax=ax, orientation='horizontal')
        cbar.ax.tick_params(labelsize=10)
        cbar.set_label('Peak Fraction', fontsize=14)

        ax.imshow(peaks.loc[self.travel_graph.index].fillna(0).values, cmap=plt.cm.rainbow)
        ax.set_yticks(np.arange(0, len(self.travel_graph.index)))
        ax.set_yticklabels(self.travel_graph.index, fontsize=10)
        ax.set_xticks(np.arange(0, peaks.shape[1], 3))
        ax.set_xticklabels(np.arange(0, peaks.shape[1], 3), fontsize=10)
        # ax.set_aspect(1)
        fig.patch.set_facecolor('#FFFFFF')


### −∗− mode : python ; −∗−
# @file EpiModel.py
# @author Bruno Goncalves
######################################################

import warnings
import networkx as nx
import numpy as np
from numpy import linalg
from numpy import random
import scipy.integrate
import pandas as pd
import matplotlib.pyplot as plt
import string
from .utils import *

class EpiModel(object):
    """Simple Epidemic Model Implementation
    
        Provides a way to implement and numerically integrate 
    """
    def __init__(self, compartments=None):
        """
        Initialize the EpiModel object
        
        Parameters:
        - compartments: list of strings, optional
            List of compartment names
        
        Returns:
        None
        """
        self.transitions = nx.MultiDiGraph()
        self.seasonality = None
        self.population = None
        self.orig_comps = None
        
        if compartments is not None:
            self.transitions.add_nodes_from([comp for comp in compartments])
    
    def add_interaction(self, source, target, agent, rate):  
        """
        Add an interaction between two compartments
        
        Parameters:
        - source: string
            Name of the source compartment
        - target: string
            Name of the target compartment
        - agent: string
            Name of the agent
        - rate: float
            Rate of the interaction
        
        Returns:
        None
        """      
        self.transitions.add_edge(source, target, agent=agent, rate=rate)        
        
    def add_spontaneous(self, source, target, rate):
        """
        Add a spontaneous transition between two compartments
        
        Parameters:
        - source: string
            Name of the source compartment
        - target: string
            Name of the target compartment
        - rate: float
            Rate of the transition
        
        Returns:
        None
        """
        self.transitions.add_edge(source, target, rate=rate)

    def add_vaccination(self, source, target, rate, start):
        """
        Add a vaccination transition between two compartments
        
        Parameters:
        - source: string
            Name of the source compartment
        - target: string
            Name of the target compartment
        - rate: float
            Rate of the vaccination
        - start: int
            Start time of the vaccination
        
        Returns:
        None
        """
        self.transitions.add_edge(source, target, rate=rate, start=start)

    def add_age_structure(self, matrix, population):
        self.contact = np.asarray(matrix)
        self.population = np.asarray(population).flatten()

        assert self.contact.shape[0] == self.contact.shape[1], "The contact matrix must be square." 

        age_groups = list(string.ascii_lowercase[:len(matrix)])
        n_ages = len(age_groups)

        model = EpiModel()
        self.orig_comps = list(self.transitions.nodes())

        for node_i, node_j, data in self.transitions.edges(data=True):
            # Interacting transition
            if "agent" in data:
                for i, age_i in enumerate(age_groups):
                    node_age_i = node_i + '_' + age_i
                    node_age_j = node_j + '_' + age_i

                    for j, age_j in enumerate(age_groups):
                        agent_age = data["agent"] + '_' + age_j

                        model.add_interaction(node_age_i, node_age_j, agent_age, data["rate"]*self.contact[i][j])

            # Spontaneous transition
            else:
                for age_i in age_groups:
                    node_age_i = node_i + '_' + age_i
                    node_age_j = node_j + '_' + age_i

                    if "start" not in data:
                        model.add_spontaneous(node_age_i, node_age_j, data["rate"])
                    else:
                        # vaccination
                        model.add_vaccination(node_age_i, node_age_j, data["rate"], data["start"])

        self.transitions = model.transitions
        
    def _new_cases(self, population, time, pos):
        """
        Internal function used by integration routine
        
        Parameters:
        - population: numpy array
            Current population of each compartment
        - time: float
            Current time
        - pos: dict
            Dictionary mapping compartment names to indices
        
        Returns:
        numpy array
            Array of new cases for each compartment
        """
        diff = np.zeros(len(pos))
        N = np.sum(population)

        if self.population is not None:
            N = {}

            for comp_i in self.transitions.nodes():
                age_group = comp_i.split('_')[-1]

                for comp_j in pos:
                    if comp_j.endswith(age_group):
                        N[comp_i] = N.get(comp_i, 0) + population[pos[comp_j]]
        
        for edge in self.transitions.edges(data=True):
            source = edge[0]
            target = edge[1]
            trans = edge[2]
            
            rate = trans['rate']*population[pos[source]]
            
            if 'start' in trans and trans['start'] >= time:
                continue

            if 'agent' in trans:
                agent = trans['agent']

                if self.population is None:
                    rate *= population[pos[agent]]/N
                else:
                    rate *= population[pos[agent]]/N[agent]

                if self.seasonality is not None:
                    curr_t = int(time)%365
                    season = float(self.seasonality[curr_t])
                    rate *= season
                
            diff[pos[source]] -= rate
            diff[pos[target]] += rate
            
        return diff
    
    def plot(self, title=None, normed=True, show=True, ax=None, **kwargs):
        """
        Convenience function for plotting
        
        Parameters:
        - title: string, optional
            Title of the plot
        - normed: bool, default=True
            Whether to normalize the values or not
        - ax: matplotlib Axes object, default=None
            The Axes object to plot to. If None, a new figure is created.
        - show: bool, default=True
            Whether to call plt.show() or not
        - kwargs: keyword arguments
            Additional arguments to pass to the plot function
        
        Returns:
        matplotlib.axes._subplots.AxesSubplot
            The plot object
        """
        try:
            if normed:
                N = self.values_.iloc[0].sum()
            else:
                N = 1

            if ax is None:
                ax = plt.gca()

            for comp in self.values_.columns:
                (self.values_[comp]/N).plot(c=epi_colors[comp[0]], **kwargs)

            ax.legend(self.values_.columns)
            ax.set_xlabel('Time')
            ax.set_ylabel('Population')
            
            if title is not None:
                ax.set_title(title)
            
            if show:
                plt.show()

            return ax
        except Exception as e:
            print(e)
            raise NotInitialized('You must call integrate() or simulate() first')
    
    def __getattr__(self, name):
        """
        Dynamic method to return the individual compartment values
        
        Parameters:
        - name: string
            Name of the compartment
        
        Returns:
        pandas.Series
            The values of the specified compartment
        """        
        if 'values_' in self.__dict__:
            return self.values_[name]
        else:
            raise AttributeError("'EpiModel' object has no attribute '%s'" % name)

    def simulate(self, timesteps, t_min=1, seasonality=None, **kwargs):
        """
        Stochastically simulate the epidemic model
        
        Parameters:
        - timesteps: int
            Number of time steps to simulate
        - t_min: int, optional
            Starting time
        - seasonality: numpy array, optional
            Array of seasonal factors
        - kwargs: keyword arguments
            Initial population of each compartment
        
        Returns:
        None
        """
        pos = {comp: i for i, comp in enumerate(self.transitions.nodes())}
        population=np.zeros(len(pos), dtype='int')

        for comp in kwargs:
            population[pos[comp]] = kwargs[comp]

        values = []
        values.append(population)

        comps = list(self.transitions.nodes)
        time = np.arange(t_min, t_min+timesteps, 1, dtype='int')

        self.seasonality = seasonality

        for t in time:
            pop = values[-1]
            new_pop = values[-1].copy()
            N = np.sum(pop)


            for comp in comps:
                trans = list(self.transitions.edges(comp, data=True))             

                prob = np.zeros(len(comps), dtype='float')

                for _, node_j, data in trans:
                    source = pos[comp]
                    target = pos[node_j]

                    rate = data['rate']

                    if 'start' in data and data['start'] >= t:
                        continue

                    if 'agent' in data:
                        agent = pos[data['agent']]
                        rate *= pop[agent]/N

                        if self.seasonality is not None:
                            curr_t = int(t)%365
                            season = float(self.seasonality[curr_t])
                            rate *= season

                    prob[target] = rate

                prob[source] = 1-np.sum(prob)

                delta = random.multinomial(pop[source], prob)
                delta[source] = 0

                changes = np.sum(delta)

                if changes == 0:
                    continue

                new_pop[source] -= changes

                for i in range(len(delta)):
                    new_pop[i] += delta[i]

            values.append(new_pop)

        values = np.array(values)
        self.values_ = pd.DataFrame(values[1:], columns=comps, index=time)
    
    def integrate(self, timesteps, t_min=1, seasonality=None, **kwargs):
        """
        Numerically integrate the epidemic model
        
        Parameters:
        - timesteps: int
            Number of time steps to integrate
        - t_min: int, optional
            Starting time
        - seasonality: numpy array, optional
            Array of seasonality values
        - kwargs: keyword arguments
            Initial population of each compartment
        
        Returns:
        None
        """        
        pos = {comp: i for i, comp in enumerate(self.transitions.nodes())}
        population=np.zeros(len(pos))

        for comp in kwargs:
            if self.population is None:
                if comp not in pos:
                    continue

                population[pos[comp]] = kwargs[comp]
            else:
                total_pop = self.population.sum()
                p = np.copy(self.population)/total_pop
                n = np.random.multinomial(kwargs[comp], p, 1)[0]

                for i, age in enumerate(string.ascii_lowercase[:len(p)]):
                    comp_age = comp + '_' + age
                    if comp_age not in pos:
                        continue

                    population[pos[comp_age]] = n[i]
        
        time = np.arange(t_min, t_min+timesteps, 1)

        self.seasonality = seasonality
        values = pd.DataFrame(scipy.integrate.odeint(self._new_cases, population, time, args=(pos,)), columns=pos.keys(), index=time)

        if self.population is None:
            self.values_ = values
        else:
            self.values_ages_ = values

            totals = values.T.copy()
            totals['key'] = totals.index.map(lambda x: '_'.join(x.split('_')[:-1]))
            totals = totals.groupby('key').sum().T
            totals.columns.name = None
            self.values_ = totals[self.orig_comps].copy()

    def single_step(self, seasonality=None, **kwargs):
        if hasattr(self, 'values_') is False:
            self.simulate(2, 1, seasonality=seasonality, **kwargs)
        else:
            old_values = self.values_.copy()
            t_curr = self.values_.index.max()
            self.simulate(2, t_curr, seasonality=seasonality, **kwargs)
            new_values = pd.concat([old_values, self.values_.iloc[[-1]]])
            self.values_ = new_values

    def __repr__(self):
        """
        Return a string representation of the EpiModel object
        
        Returns:
        string
            String representation of the EpiModel object
        """
        text = 'Epidemic Model with %u compartments and %u transitions:\n\n' % \
              (self.transitions.number_of_nodes(), 
               self.transitions.number_of_edges())
        
        for edge in self.transitions.edges(data=True):
            source = edge[0]
            target = edge[1]
            trans = edge[2]
            
            rate = trans['rate']

            if 'agent' in trans:
                agent = trans['agent']
                text += "%s + %s = %s %f\n" % (source, agent, target, rate)
            elif 'start' in trans:
                start = trans['start']
                text+="%s -> %s %f starting at %s days\n" % (source, target, rate, start)
            else:
                text+="%s -> %s %f\n" % (source, target, rate)
        
        R0 = self.R0()

        if R0 is not None:
            text += "\nR0=%1.2f" % R0

        return text

    def _get_active(self):
        active = set()

        for node_i, node_j, data in self.transitions.edges(data=True):
            if "agent" in data:
                active.add(data['agent'])
            else:
                active.add(node_i)

        return active

    def _get_susceptible(self):
        susceptible = set([node for node, deg in self.transitions.in_degree() if deg==0])

        if len(susceptible) == 0:
            for node_i, node_j, data in self.transitions.edges(data=True):
                if "agent" in data:
                    susceptible.add(node_i)

        return susceptible

    def _get_infections(self):
        inf = {}

        for node_i, node_j, data in self.transitions.edges(data=True):
            if "agent" in data:
                agent = data['agent']

                if agent not in inf:
                    inf[agent] = {}

                if node_i not in inf[agent]:
                    inf[agent][node_i] = {}

                inf[agent][node_i]['target'] = node_j
                inf[agent][node_i]['rate'] = data['rate']

        return inf

    def draw_model(self, ax=None, show=True):
        """
        Plot the model structure

        - ax: matplotlib Axes object, default=None
            The Axes object to plot to. If None, a new figure is created.
        - show: bool, default=True
            Whether to call plt.show() or not
        """
        try:
            from networkx.drawing.nx_agraph import graphviz_layout
            pos=graphviz_layout(self.transitions, prog='dot', args='-Grankdir="LR"')
        except:
            pos=nx.layout.spectral_layout(self.transitions)

        node_colors = []

        for node in self.transitions.nodes():
            node_colors.append(epi_colors[node[0]])

        edge_labels = {}

        for node_i, node_j, data in self.transitions.edges(data=True):
            edge = (node_i, node_j)

            if "agent" in data:
                if edge not in edge_labels:
                    edge_labels[edge] = data["agent"]
                else:
                    edge_labels[edge] = edge_labels[edge] + "+" + data["agent"]
            else:
                edge_labels[edge] = ""


        if ax is None:
            fig, ax = plt.subplots(1, figsize=(10, 2))

        nx.draw(self.transitions, pos, with_labels=True, arrows=True, node_shape='H', 
        font_color='k', node_color=node_colors, node_size=1000, ax=ax)
        nx.draw_networkx_edge_labels(self.transitions, pos, edge_labels=edge_labels, ax=ax)

        if show:
            plt.show()

    def R0(self):
        """
        Return the value of the basic reproductive ratio, $R_0$, for the model as defined

        The calculation is completely generic as it uses the Next-Generation matrix approach
        defined in J. R. Soc Interface 7, 873 (2010)

        Returns:
        R0 - the value of the largest eigenvalue of the next generation matrix
        """

        infected = set()

        susceptible = self._get_susceptible()

        for node_i, node_j, data in self.transitions.edges(data=True):
            if "agent" in data:
                infected.add(data['agent'])
                infected.add(node_j)


        infected = sorted(infected)
        N_infected = len(infected)

        F = np.zeros((N_infected, N_infected), dtype='float')
        V = np.zeros((N_infected, N_infected), dtype='float')

        pos = dict(zip(infected, np.arange(N_infected)))

        try:
            for node_i, node_j, data in self.transitions.edges(data=True):
                rate = data['rate']

                if "agent" in data:
                    target = pos[node_j]
                    agent = pos[data['agent']]

                    if node_i in susceptible:
                        F[target, agent] = rate
                elif "start" in data:
                    continue
                else:
                    source = pos[node_i]

                    V[source, source] += rate

                    if node_j in pos:
                        target = pos[node_j]
                        V[target, source] -= rate
        
            eig, v = linalg.eig(np.dot(F, linalg.inv(V)))
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                return np.real(eig.max())
        except:
            return None

    def __getitem__(self, key):
        if type(key) != type([]):
            key_check = set([key])
        else:
            key_check = set(key)

        if len(key_check & set(self.values_.columns)) > 0:
            return self.values_[key]
        elif len(key_check & set(self.values_ages_.columns)) > 0:
            return self.values_ages_[key]
        else:
            return None



from collections import defaultdict

class NotInitialized(Exception):
    pass

epi_colors = defaultdict(lambda :'#f39019')
epi_colors['S'] = '#51a7f9'
epi_colors['E'] = '#f9e351'
epi_colors['I'] =  '#cf51f9'
epi_colors['R'] = '#70bf41'
epi_colors['D'] = '#8b8b8b'


import unittest
import pandas as pd
from epidemik import MetaEpiModel
from epidemik.utils import NotInitialized

class MetaEpiModelTestCase(unittest.TestCase):
	def setUp(self):
		self.travel = pd.DataFrame({'A': [0.99, 0.1], 'B':[0.01, 0.9]}, index=["A", "B"])
		self.population = pd.DataFrame({'Population':[100_000, 10_000]}, index=["A", "B"])

		self.SIR = MetaEpiModel(self.travel, self.population)
		self.beta = 0.3
		self.mu = 0.1
		self.SIR.add_interaction('S', 'I', 'I', self.beta)
		self.SIR.add_spontaneous('I', 'R', self.mu)

	def test_number_populations(self):
		self.assertEqual(self.SIR.travel_graph.shape[0], 2)

	def test_R0(self):
		self.assertEqual(self.SIR.R0(), 3.0, 'incorrect R0')

	def test_simulate(self):
		with self.assertRaises(NotInitialized) as _:
			self.SIR.simulate(10)

	def test_initialize_populations(self):
		self.SIR._initialize_populations('S')
		self.assertEqual(self.SIR.compartments_['S'].sum(), self.population['Population'].sum())

	def test_travel(self):
		self.SIR._initialize_populations('S')
		new_compartments = self.SIR._run_travel(self.SIR.compartments_, self.travel)

		self.assertEqual(new_compartments.sum().sum(), self.population['Population'].sum())

	def test_integrate(self):
		with self.assertRaises(NotImplementedError) as _:
			self.SIR.integrate()